/*
 * FieldInit.cpp
 *
 *  Created on: Jul 27, 2009
 *      Author: zdevito
 */

#include "MPI/FieldInit.h"
#include "MPI/MPIUtil.h"


namespace LisztMPIPrivate {

FieldInit::FieldInit() : owner_map(NULL) {}

void FieldInit::init(MPI_Comm c, LisztPrivate::OwnershipMap * map) {
    comm = c;
    owner_map = map;
}

template<typename ElemType>
void FieldInit::initField(FieldInfo * fi) {
    assert(owner_map);
    field_info = fi;
    field_info->registered_readers.reset(new CommPair);
    field_info->registered_writers.reset(new CommPair);
    registerGhosts<ElemType>();
    field_info = NULL;
}

template<typename ElemType>
void FieldInit::registerGhosts() {
    registerComm<ElemType>(field_info->ghost_reads,&field_info->registered_readers->recv,&field_info->registered_readers->send,comm);
    registerComm<ElemType>(field_info->ghost_writes,&field_info->registered_writers->send,&field_info->registered_writers->recv,comm);
}

template<typename ElemType>
void FieldInit::registerComm(const FieldInit::GhostSet & commset, CommInfo * from, CommInfo * to, MPI_Comm comm) {
    using namespace LisztPrivate::ElemTypes;
    //establish a communication pathway to keep fields up to date
    //commset is the set of ghost elements for which we are communicating values for this field, represented by global ids
    //we will look up the owners of each of these elements and then notify them that we are reading/writing

    //if this is a Writer->Owner update
    //from will be a communicator representing the messages sent from this node to the owners
    //to will be a communicator representing messages received by this node from the writers

    //if this is a Owner->Reader update
    //from will be a communicator representing the messages to be received from the owners
    //to will be a communicator representing the messages to be sent to the readers
    int size;
    MPI_Comm_size(comm,&size);
    int rank;
    MPI_Comm_rank(comm,&rank);
    size_t mpi_size = size;

   //commset.size() is the total number of ghosts for this field
   int * send_buf = new int[commset.size()];

   //calcuate the size of each message
   int * send_sizes = new int[mpi_size];
   std::fill_n(send_sizes,mpi_size,0);


   for(GhostSet::iterator it = commset.begin(), end = commset.end();
        it != end;
        it++) {
       rank_type r = owner_map->getOwnerFor(ElemRef<ElemType>(*it));
       assert(r >= 0 && r < static_cast<rank_type>(mpi_size) && r != rank);
       send_sizes[static_cast<size_t>(r)]++;
    }
   int * send_disp = new int[mpi_size];
   if(MPIU_GetDispFromSize(send_sizes,mpi_size,send_disp) != (int) commset.size())
   		assert(!"commset is inconsistent");

   //communicate to get the recv size
   int * recv_sizes = new int[mpi_size];

   MPI_Alltoall(send_sizes,1,MPI_INT,recv_sizes,1,MPI_INT,comm);

   int * recv_disp = new int[mpi_size];
   int recv_total_size = MPIU_GetDispFromSize(recv_sizes,mpi_size,recv_disp);

   int * recv_buf = new int [recv_total_size];

   //fill the send_buf with the messages
   //we send the list of global ids that we read/write to their owner
   int * send_pos = new int[mpi_size];
   std::fill_n(send_pos,mpi_size,0);
   for(GhostSet::iterator it = commset.begin(), end = commset.end();
        it != end;
        it++) {
       rank_type r = owner_map->getOwnerFor(ElemRef<ElemType>(*it));
       assert(r >= 0 && r < static_cast<rank_type>(mpi_size));
       send_buf[send_disp[static_cast<size_t>(r)] + send_pos[static_cast<size_t>(r)]++] = *it;
    }
   for(size_t i = 0; i < mpi_size; i++)
       assert(send_pos[i] == send_sizes[i]);
   delete [] send_pos;

   MPI_Alltoallv(send_buf,send_sizes,send_disp,MPI_INT,recv_buf,recv_sizes,recv_disp,MPI_INT,comm);
   //recv_buf now holds the messages that will be received from the readers/writers


   //we use these MPI buffers to initialize the Comm objects, which take OWNERSHIP send_buf/recv_buf
   //but only use send_sizes, and send_disp for initialization
   from->init(reinterpret_cast<id_type*>(send_buf),send_sizes,send_disp,mpi_size);
   to->init(reinterpret_cast<id_type*>(recv_buf),recv_sizes,recv_disp,mpi_size);


   delete [] send_sizes;
   delete [] send_disp;

   delete [] recv_disp;
   delete [] recv_sizes;
}


#define MAKE_TEMPLATES(elem_type) \
template \
void FieldInit::registerComm<elem_type>(const GhostSet & commset, CommInfo * from, CommInfo * to, MPI_Comm comm); \
template \
void FieldInit::registerGhosts<elem_type>(); \
template \
void FieldInit::initField<elem_type>(FieldInfo * fi);

MAKE_TEMPLATES(LisztPrivate::VertexType)
MAKE_TEMPLATES(LisztPrivate::EdgeType)
MAKE_TEMPLATES(LisztPrivate::FaceType)
MAKE_TEMPLATES(LisztPrivate::CellType)

#undef MAKE_TEMPLATES

}  // namespace LisztMPIPrivate
