/*
 * Liszt.cpp
 *
 *  Created on: Jul 27, 2009
 *      Author: zdevito
 */

#include "MPI/LisztMPI.h"
#include "MPI/MPIUtil.h"

namespace LisztMPIPrivate {

using namespace CRSMesh;
using namespace KernelDescription;


LisztMPI::LisztMPI() : is_prepared(false) {

}
void LisztMPI::init(int argc, char ** argv) {
    MPI_Init(&argc,&argv);
    PETSC_DO(PetscInitialize(&argc,&argv,(char*)0,"Liszt"));
    int rank,size;
    MPI_Comm_dup(MPI_COMM_WORLD,&liszt_comm);
    MPI_Comm_size(liszt_comm,&size);
    MPI_Comm_rank(liszt_comm,&rank);
    mpi_rank = static_cast<rank_type>(rank);
    mpi_size = static_cast<size_t>(size);
}
void LisztMPI::bindTopology(::std::tr1::shared_ptr<MeshTopology> top) {
    topology = top;
}

template<typename T>
 void LisztMPI::addSet(const ElemRefSet<T> & set, OwnershipRequest * req) const {
    typedef typename ElemRefSet<T>::const_iterator iterator;
    for(iterator it = set.begin(), end = set.end(); it != end; it++) {
        ElemRef<T> e = *it;
        //if(local.hasFields(e)) {
            req->addRequest(e);
        //}
    }
}

 void LisztMPI::createOwnershipMap(const Partitioner & partitioner, OwnershipMap * map) const {
    /*OwnershipRequest request(*topology,local.numberOfGhostsWithFields<VertexType>(),
                                       local.numberOfGhostsWithFields<EdgeType>(),
                                       local.numberOfGhostsWithFields<FaceType>(),
                                       local.numberOfGhostsWithFields<CellType>());*/
    OwnershipRequest request(*topology,local.numberGhosts<VertexType>(),
                                       local.numberGhosts<EdgeType>(),
                                       local.numberGhosts<FaceType>(),
                                       local.numberGhosts<CellType>());

    addSet(local.ghost<VertexType>(),&request);
    addSet(local.ghost<EdgeType>(),&request);
    addSet(local.ghost<FaceType>(),&request);
    addSet(local.ghost<CellType>(),&request);
    map->init(topology);
    partitioner.requestOwnership(request,map);
}
struct EstablishFieldComm : public FieldDescriptionAdaptor {
    EstablishFieldComm(MPI_Comm c, OwnershipMap * owner_map, FieldDescriptorMap<FieldInfo> * fi)
    : field_info(fi) {
        init.init(c,owner_map);
    }
    virtual void visitVertexFieldDescriptor(VertexFieldDescriptor * obj) {
        init.initField<VertexType>( &(*field_info)[obj] );
    }
    virtual void visitEdgeFieldDescriptor(EdgeFieldDescriptor * obj) {
        init.initField<EdgeType>( &(*field_info)[obj] );
    }
    virtual void visitFaceFieldDescriptor(FaceFieldDescriptor * obj) {
        init.initField<FaceType>( &(*field_info)[obj] );
    }
    virtual void visitCellFieldDescriptor(CellFieldDescriptor * obj) {
        init.initField<CellType>( &(*field_info)[obj] );
    }
private:
    FieldInit init;
    FieldDescriptorMap<FieldInfo> * field_info;
};
 void LisztMPI::bindDescription(::std::tr1::shared_ptr<KernelDescription::KernelDescriptor> k) {
    kernel = k;
    kernel->buildSetDescriptorMap(&set_map);
    kernel->buildFieldDescriptorMap(&field_info);
 }
 void LisztMPI::prepare() {
     assert(!is_prepared);
    if(kernel.get() == NULL) {
        cout  << "description not bound!" << endl;
        assert(false);
    }
    if(topology.get() == NULL) {
        cout << "topology not bound" << endl;
        assert(false);
    }
    Partitioner partitioner(mpi_rank,mpi_size,liszt_comm);
    partitioner.partition(*topology,*kernel,&local);
    partitioner.printDist();
    local.finishedOwned();
    cout << "partitioning complete" << endl;
    //kernel->bindLocalElements(&local);

    DiscoverGhosts discoverghosts;
    discoverghosts.init(topology.get(),kernel.get());
    discoverghosts.discover(&local,&set_map,&field_info);

    partitioner.printDist();
    local.finishedGhost();
    cout << "ghost cell discovery complete" << endl;

    createOwnershipMap(partitioner,&owner_map);
    //kernel->bindOwnershipMap(&owner_map);
    cout << "ghost ownership determined" << endl;
    //kernel->registerAllGhostFields(liszt_comm);
    EstablishFieldComm efc(liszt_comm,&owner_map,&field_info);
    kernel->visitFieldDescriptors(&efc);
    partitioner.printDist();
    cout << "field communication established." << endl;
    buildPartitionedGIDs();
    is_prepared = true;
}

template<typename elem_type>
void LisztMPI::buildPGIDForElem() {
	int * send_sizes = new int[mpi_size];
	std::fill(send_sizes, send_sizes + mpi_size, 0);
	ElemRefSet<elem_type> ghost = local.ghost<elem_type>();
	id_type * owners = new id_type[ghost.size()];
	{
		int i = 0;
		for(typename ElemRefSet<elem_type>::const_iterator it = ghost.begin(), end = ghost.end();
		   it != end;
		   it++) {
		   //cout << "I have " << elem_type::value << " ghost: " << it->ID() << endl;
			rank_type r = owner_map.getOwnerFor(*it);
			owners[i++] = r;
			send_sizes[r]++;
		}
	}
	int * send_disp = new int[mpi_size];
	int * send_off = new int[mpi_size];
	std::fill(send_off,send_off + mpi_size,0);
	
	int send_total_size = MPIU_GetDispFromSize(send_sizes,mpi_size,send_disp);
	/*cout << "Send Sizes: " << endl;
	for(int i = 0; i < mpi_size; i++) {
		cout << i << " " << send_sizes[i] << endl;
	}*/
	int * recv_sizes = new int[mpi_size];
	MPI_Alltoall(send_sizes,1, MPI_INT, recv_sizes, 1, MPI_INT, liszt_comm);
	/*cout << "Recieve Sizes: " << endl;
	for(int i = 0; i < mpi_size; i++) {
		cout << i << " " << recv_sizes[i] << endl;
	}*/

	int * recv_disp = new int[mpi_size];
	int recv_total_size = MPIU_GetDispFromSize(recv_sizes,mpi_size,recv_disp);
	
	int * send_buf = new int[send_total_size];
	int * recv_buf = new int[recv_total_size];
	{
		int i = 0;
		for(typename ElemRefSet<elem_type>::const_iterator it = ghost.begin(), end = ghost.end();
		    it != end;
		    it++) {
			rank_type r = owners[i];
			int id_location = send_disp[r] + send_off[r];
			send_buf[id_location] = it->ID();
			owners[i] = id_location;
			send_off[r]++;
			i++;
		}
	}
	for(unsigned int i = 0; i < mpi_size; i++) {
		assert(send_sizes[i] == send_off[i]);
	}
	MPI_Alltoallv(send_buf,send_sizes,send_disp,MPI_INT,recv_buf,recv_sizes,recv_disp,MPI_INT,liszt_comm);
	
	for(int i = 0; i < recv_total_size; i++) {
		recv_buf[i] = local_id_offset[elem_type::value] + local.getLocalID(ElemRef<elem_type>(recv_buf[i]));
	}
	
	MPI_Alltoallv(recv_buf,recv_sizes,recv_disp,MPI_INT,send_buf,send_sizes,send_disp,MPI_INT,liszt_comm);
	{
		int i = 0;
		for(typename ElemRefSet<elem_type>::const_iterator it = ghost.begin(), end = ghost.end();
		    it != end;
		    it++) {
			owners[i] = send_buf[owners[i]];
			i++;
		}
	}
	local_ghost_to_partitioned_gid[elem_type::value] = owners;
	delete [] send_sizes;
	delete [] send_disp;
	delete [] send_off;
	delete [] recv_sizes;
	delete [] recv_disp;
	delete [] send_buf;
	delete [] recv_buf;
}
void LisztMPI::buildPartitionedGIDs() {
	num_local_elems[VertexType::value] = local.numberOwned<VertexType>();
	num_local_elems[EdgeType::value] = local.numberOwned<EdgeType>();
	num_local_elems[FaceType::value] = local.numberOwned<FaceType>();
	num_local_elems[CellType::value] = local.numberOwned<CellType>();
	
	size_t (*local_sizes)[ELEM_TYPE_SIZE] = new size_t[mpi_size][ELEM_TYPE_SIZE];
	//TODO(zach): would be faster to do a parallel scan ...
	MPI_Allgather(num_local_elems,sizeof(size_t)*ELEM_TYPE_SIZE,MPI_BYTE,local_sizes,sizeof(size_t) * ELEM_TYPE_SIZE,MPI_BYTE,liszt_comm);
	//calc local offset
	std::fill(local_id_offset,local_id_offset + ELEM_TYPE_SIZE,0);
	for(int i = 0; i < mpi_rank; i++) {
		for(size_t j = 0; j < ELEM_TYPE_SIZE; j++)
			local_id_offset[j] += local_sizes[i][j];
	}
	delete [] local_sizes;
	
	cout << mpi_rank << ": offset is " << local_id_offset[0] << " " << local_id_offset[1] << " " << local_id_offset[2] << " " << local_id_offset[3] << endl;
	
	buildPGIDForElem<VertexType>();
	buildPGIDForElem<EdgeType>();
	buildPGIDForElem<FaceType>();
	buildPGIDForElem<CellType>();
}

void LisztMPI::finalize() {

    MPI_Finalize();
}
void LisztMPI::redirectToLog(const char * base) {
    ostringstream ss;
    ss << base << "_" << mpi_rank << ".log";
    int file = fileno(fopen(ss.str().c_str(),"a"));
    dup2(file,1);
    dup2(file,2);
}
} //namespace LisztMPIPrivate
